@@ -27,13 +27,15 @@ Vinay Sajip to make use of the subprocess module (Steve's version uses os.fork()
 and so does not work on Windows). Renamed to gnupg.py to avoid confusion with
 the previous versions.
 
-Modifications Copyright (C) 2008-2023 Vinay Sajip. All rights reserved.
+Modifications Copyright (C) 2008-2024 Vinay Sajip. All rights reserved.
 
 For the full documentation, see https://docs.red-dove.com/python-gnupg/ or
 https://gnupg.readthedocs.io/
 """
 
 import codecs
+from datetime import datetime
+from email.utils import parseaddr
 from io import StringIO
 import logging
 import os
@@ -43,9 +45,9 @@ from subprocess import Popen, PIPE
 import sys
 import threading
 
-__version__ = '0.5.1'
+__version__ = '0.5.3'
 __author__ = 'Vinay Sajip'
-__date__ = '$22-Jul-2023 16:36:40$'
+__date__ = '$20-Sep-2024 17:22:27$'
 
 STARTUPINFO = None
 if os.name == 'nt':  # pragma: no cover
@@ -135,8 +137,9 @@ def no_quote(s):
     return s
 
 
-def _copy_data(instream, outstream):
+def _copy_data(instream, outstream, buffer_size):
     # Copy one stream to another
+    assert buffer_size > 0
     sent = 0
     if hasattr(sys.stdin, 'encoding'):
         enc = sys.stdin.encoding
@@ -146,7 +149,7 @@ def _copy_data(instream, outstream):
         # See issue #39: read can fail when e.g. a text stream is provided
         # for what is actually a binary file
         try:
-            data = instream.read(1024)
+            data = instream.read(buffer_size)
         except Exception:  # pragma: no cover
             logger.warning('Exception occurred while reading', exc_info=1)
             break
@@ -170,8 +173,9 @@ def _copy_data(instream, outstream):
     logger.debug('closed output, %d bytes sent', sent)
 
 
-def _threaded_copy_data(instream, outstream):
-    wr = threading.Thread(target=_copy_data, args=(instream, outstream))
+def _threaded_copy_data(instream, outstream, buffer_size):
+    assert buffer_size > 0
+    wr = threading.Thread(target=_copy_data, args=(instream, outstream, buffer_size))
     wr.daemon = True
     logger.debug('data copier: %r, %r, %r', wr, instream, outstream)
     wr.start()
@@ -210,7 +214,7 @@ def _make_binary_stream(s, encoding):
 
 class StatusHandler(object):
     """
-    The base class for handling status messages from gpg.
+    The base class for handling status messages from `gpg`.
     """
 
     def __init__(self, gpg):
@@ -998,6 +1002,55 @@ class Sign(StatusHandler, TextHandler):
             logger.debug('message ignored: %s, %s', key, value)
 
 
+class AutoLocateKey(StatusHandler):
+    """
+    This class handles status messages during key auto-locating.
+    fingerprint: str
+    key_length: int
+    created_at: date
+    email: str
+    email_real_name: str
+    """
+
+    def __init__(self, gpg):
+        StatusHandler.__init__(self, gpg)
+        self.fingerprint = None
+        self.type = None
+        self.created_at = None
+        self.email = None
+        self.email_real_name = None
+
+    def handle_status(self, key, value):
+        if key == "IMPORTED":
+            _, email, display_name = value.split()
+
+            self.email = email
+            self.email_real_name = display_name[1:-1]
+        elif key == "KEY_CONSIDERED":
+            self.fingerprint = value.strip().split()[0]
+
+    def pub(self, args):
+        """
+        Internal method to handle the 'pub' status message.
+        `pub` message contains the fingerprint of the public key, its type and its creation date.
+        """
+        pass
+
+    def uid(self, args):
+        self.created_at = datetime.fromtimestamp(int(args[5]))
+        raw_email_content = args[9]
+        email, real_name = parseaddr(raw_email_content)
+        self.email = email
+        self.email_real_name = real_name
+
+    def sub(self, args):
+        self.key_length = int(args[2])
+
+    def fpr(self, args):
+        # Only store the first fingerprint
+        self.fingerprint = self.fingerprint or args[9]
+
+
 VERSION_RE = re.compile(r'^cfg:version:(\d+(\.\d+)*)'.encode('ascii'))
 HEX_DIGITS_RE = re.compile(r'[0-9a-f]+$', re.I)
 PUBLIC_KEY_RE = re.compile(r'gpg: public key is (\w+)')
@@ -1011,6 +1064,8 @@ class GPG(object):
 
     decode_errors = 'strict'
 
+    buffer_size = 16384  # override in instance if needed
+
     result_map = {
         'crypt': Crypt,
         'delete': DeleteResult,
@@ -1025,6 +1080,7 @@ class GPG(object):
         'trust': TrustResult,
         'verify': Verify,
         'export': ExportResult,
+        'auto-locate-key': AutoLocateKey,
     }
     "A map of GPG operations to result object types."
 
@@ -1043,7 +1099,7 @@ class GPG(object):
             gpgbinary (str): A pathname for the GPG binary to use.
 
             gnupghome (str): A pathname to where we can find the public and private keyrings. The default is
-                             whatever gpg defaults to.
+                             whatever `gpg` defaults to.
 
             keyring (str|list): The name of alternative keyring file to use, or a list of such keyring files. If
                                 specified, the default keyring is not used.
@@ -1111,7 +1167,7 @@ class GPG(object):
         """
         Make a list of command line elements for GPG. The value of ``args``
         will be appended. The ``passphrase`` argument needs to be True if
-        a passphrase will be sent to GPG, else False.
+        a passphrase will be sent to `gpg`, else False.
 
         Args:
             args (list[str]): A list of arguments.
@@ -1189,11 +1245,12 @@ class GPG(object):
                 result.handle_status(keyword, value)
         result.stderr = ''.join(lines)
 
-    def _read_data(self, stream, result, on_data=None):
+    def _read_data(self, stream, result, on_data=None, buffer_size=1024):
         # Read the contents of the file from GPG's stdout
+        assert buffer_size > 0
         chunks = []
         while True:
-            data = stream.read(1024)
+            data = stream.read(buffer_size)
             if len(data) == 0:
                 if on_data:
                     on_data(data)
@@ -1224,7 +1281,7 @@ class GPG(object):
         rr.start()
 
         stdout = process.stdout
-        dr = threading.Thread(target=self._read_data, args=(stdout, result, self.on_data))
+        dr = threading.Thread(target=self._read_data, args=(stdout, result, self.on_data, self.buffer_size))
         dr.daemon = True
         logger.debug('stdout reader: %r', dr)
         dr.start()
@@ -1279,13 +1336,15 @@ class GPG(object):
                 stdin = codecs.getwriter(self.encoding)(p.stdin)
             else:
                 stdin = p.stdin
+            writer = None  # See issue #237
             if passphrase:
                 _write_passphrase(stdin, passphrase, self.encoding)
-            writer = _threaded_copy_data(fileobj, stdin)
+            writer = _threaded_copy_data(fileobj, stdin, self.buffer_size)
             self._collect_output(p, result, writer, stdin)
             return result
         finally:
-            writer.join(0.01)
+            if writer:
+                writer.join(0.01)
             if fileobj is not fileobj_or_path:
                 fileobj.close()
 
@@ -1337,8 +1396,8 @@ class GPG(object):
 
     def is_valid_passphrase(self, passphrase):
         """
-        Confirm that the passphrase doesn't contain newline-type characters - it is passed in a pipe to gpg, and so not
-        checking could lead to spoofing attacks by passing arbitrary text after passphrase and newline.
+        Confirm that the passphrase doesn't contain newline-type characters - it is passed in a pipe to `gpg`,
+        and so not checking could lead to spoofing attacks by passing arbitrary text after passphrase and newline.
 
         Args:
             passphrase (str): The passphrase to test.
@@ -1407,7 +1466,7 @@ class GPG(object):
             stdin = p.stdin
             if passphrase:
                 _write_passphrase(stdin, passphrase, self.encoding)
-            writer = _threaded_copy_data(fileobj, stdin)
+            writer = _threaded_copy_data(fileobj, stdin, self.buffer_size)
         except IOError:  # pragma: no cover
             logging.exception('error writing message')
             writer = None
@@ -1606,7 +1665,7 @@ class GPG(object):
         .. note:: Passphrases
 
            Since GnuPG 2.1, you can't delete secret keys without providing a passphrase. However, if you're expecting
-           the passphrase to go to gpg via pinentry, you should specify expect_passphrase=False. (It's only checked
+           the passphrase to go to `gpg` via pinentry, you should specify expect_passphrase=False. (It's only checked
            for GnuPG >= 2.1).
         """
         if passphrase and not self.is_valid_passphrase(passphrase):  # pragma: no cover
@@ -1651,7 +1710,7 @@ class GPG(object):
                     expect_passphrase=True,
                     output=None):
         """
-        Export the indicated keys. A 'keyid' is anything gpg accepts.
+        Export the indicated keys. A 'keyid' is anything `gpg` accepts.
 
         Args:
             keyids (str|list[str]): A single keyid or a list of them.
@@ -1671,7 +1730,7 @@ class GPG(object):
         .. note:: Passphrases
 
            Since GnuPG 2.1, you can't export secret keys without providing a passphrase. However, if you're expecting
-           the passphrase to go to gpg via pinentry, you should specify expect_passphrase=False. (It's only checked
+           the passphrase to go to `gpg` via pinentry, you should specify expect_passphrase=False. (It's only checked
            for GnuPG >= 2.1).
         """
         if passphrase and not self.is_valid_passphrase(passphrase):  # pragma: no cover
@@ -1864,6 +1923,30 @@ class GPG(object):
                 getattr(result, keyword)(L)
         return result
 
+    def auto_locate_key(self, email, mechanisms=None, **kwargs):
+        """
+        Auto locate a public key by `email`.
+
+        Args:
+            email (str): The email address to search for.
+            mechanisms (list[str]): A list of mechanisms to use. Valid mechanisms can be found
+            here https://www.gnupg.org/documentation/manuals/gnupg/GPG-Configuration-Options.html
+            under "--auto-key-locate". Default: ['wkd', 'ntds', 'ldap', 'cert', 'dane', 'local']
+        """
+        mechanisms = mechanisms or ['wkd', 'ntds', 'ldap', 'cert', 'dane', 'local']
+
+        args = ['--auto-key-locate', ','.join(mechanisms), '--locate-keys', email]
+
+        result = self.result_map['auto-locate-key'](self)
+
+        if 'extra_args' in kwargs:
+            args.extend(kwargs['extra_args'])
+
+        process = self._open_subprocess(args)
+        self._collect_output(process, result, stdin=process.stdin)
+        self._decode_result(result)
+        return result
+
     def gen_key(self, input):
         """
         Generate a key; you might use `gen_key_input()` to create the input.
@@ -1880,7 +1963,7 @@ class GPG(object):
 
     def gen_key_input(self, **kwargs):
         """
-        Generate `--gen-key` input  (see gpg documentation in DETAILS).
+        Generate `--gen-key` input  (see `gpg` documentation in DETAILS).
 
         Args:
             kwargs (dict): A list of keyword arguments.
@@ -2114,7 +2197,7 @@ class GPG(object):
             args.extend(extra_args)
         result = self.result_map['crypt'](self)
         self._handle_io(args, fileobj_or_path, result, passphrase, binary=True)
-        logger.debug('decrypt result[:100]: %r', result.data[:100])
+        # logger.debug('decrypt result[:100]: %r', result.data[:100])
         return result
 
     def get_recipients(self, message, **kwargs):
